use relm4::{
    SimpleComponent,
    adw::{self, prelude::*},
    gtk::glib::clone,
};

pub struct Dialog {
    // toast overlay looks like an odd choice but i think it's fitting considering it acts as a
    // "box" for the main content
    window: adw::ToastOverlay,
    visible: bool,
    heading: String,
    body: String,
}

#[derive(Debug)]
pub enum DialogInput {
    Show { heading: String, body: String },
    Dismiss,
}

pub struct DialogWidgets {
    dialog: adw::AlertDialog,
}

impl SimpleComponent for Dialog {
    type Init = adw::ToastOverlay;
    type Root = adw::AlertDialog;
    type Widgets = DialogWidgets;
    type Input = DialogInput;
    type Output = ();

    fn init_root() -> Self::Root {
        let dialog = adw::AlertDialog::builder().close_response("ok").build();
        dialog.add_response("ok", "OK");
        dialog
    }

    fn init(
        window: Self::Init,
        root: Self::Root,
        sender: relm4::ComponentSender<Self>,
    ) -> relm4::ComponentParts<Self> {
        let model = Self {
            window,
            visible: false,
            heading: String::new(),
            body: String::new(),
        };

        root.connect_response(
            None,
            clone!(
                #[strong]
                sender,
                move |_, _| sender.input(DialogInput::Dismiss)
            ),
        );

        let widgets = DialogWidgets { dialog: root };

        relm4::ComponentParts { model, widgets }
    }

    fn update(&mut self, message: Self::Input, _sender: relm4::ComponentSender<Self>) {
        match message {
            DialogInput::Show { heading, body } => {
                self.heading = heading;
                self.body = body;
                self.visible = true;
            }
            DialogInput::Dismiss => self.visible = false,
        }
    }

    fn update_view(&self, widgets: &mut Self::Widgets, _sender: relm4::ComponentSender<Self>) {
        widgets.dialog.set_heading(Some(&self.heading));
        widgets.dialog.set_body(&self.body);

        if self.visible {
            widgets.dialog.present(Some(&self.window));
        }
    }
}

pub struct Toast {
    root: adw::ToastOverlay,
    visible: bool,
    text: String,
}

#[derive(Debug)]
pub enum ToastInput {
    Show(String),
    Copy,
    Dismiss,
}

#[relm4::component(pub)]
impl SimpleComponent for Toast {
    type Input = ToastInput;
    type Output = ();
    type Init = adw::ToastOverlay;

    view! {
        #[name = "toast"]
        adw::Toast {
            #[watch]
            set_title: &model.text,
            set_button_label: Some("Copy"),
            connect_button_clicked => ToastInput::Copy,
            connect_dismissed => ToastInput::Dismiss,
        }
    }

    fn init(
        init: Self::Init,
        root: Self::Root,
        _sender: relm4::ComponentSender<Self>,
    ) -> relm4::ComponentParts<Self> {
        let model = Self {
            root: init,
            visible: false,
            text: String::new(),
        };

        let widgets = view_output!();
        relm4::ComponentParts { model, widgets }
    }

    fn update(&mut self, message: Self::Input, _sender: relm4::ComponentSender<Self>) {
        match message {
            ToastInput::Show(text) => {
                self.visible = true;
                self.text = text;
            }
            ToastInput::Copy => {
                if let Err(e) =
                    arboard::Clipboard::new().and_then(|mut c| c.set_text(self.text.clone()))
                {
                    tracing::error!("could not copy url to clipboard: {e}");
                    self.text = "Could not copy".into();
                } else {
                    self.visible = false;
                }
            }
            ToastInput::Dismiss => self.visible = false,
        }
    }

    fn post_view() {
        if model.visible {
            model.root.add_toast(widgets.toast.clone());
        }
    }
}
